package agt

import (
	"ia04/comsoc"
	"net/http"
	"sync"
	"encoding/json"
	"bytes"
	"fmt"
	"time"
	"log"
)

type RequestVote struct {
	voteID 		string					`json:"vote-id"`
	prefs			[]Alternative 	`json:"prefs"`
	agentID   int 						`json:"agent-id"`
	options   []Alternative		`json:"options"`
}

type RequestNewBallot struct {
	rule       string					`json:"rule"`
	deadline   string					`json:"deadline"`
	voterIDs   []string				`json:"voter-ids"`
	alts       int					  `json:"#alts"`
}

type RequestResult struct {
	ballotID  string					`json:"ballot-id"`
}

type ResponseNewBallot struct {
	ballotID  string					`json:"ballot-id"`
	// code      int
}

// type ResponseVote struct {
// 	code      int
// }

type ResponseResult struct {
	winner    int             `json:"winner"`
	ranking   []Alternative		`json:"ranking"`
  // code			int
}



type BallotID int

type RestBallotAgent struct {
	sync.Mutex
	ID   BallotID
	addr string		// the only non-varying value in one
	 							// server instance, the rest is change
								// every RequestNewBallot
	pfl  Profile
	rule string
	isOpen bool	// open: 1 / closed: 0
	voterIDs []string
	alts []Alternative
}

type Alternative comsoc.Alternative
type Profile comsoc.Profile

type RestBallotI interface {
	checkMethod(method string, w http.ResponseWriter, r *http.Request) bool
	decodeRequestNewBallot(r *http.Request) (req RequestNewBallot, err error)
	decodeRequestVote(r *http.Request) (req RequestVote, err error)
	decodeRequestResult(r *http.Request) (req RequestResult, err error)
}  // je ne sais pas trop comment utiliser l'interface je te laisse si tu préfères

func NewRestBallotAgent(addr string, emptyProfile Profile) *RestBallotAgent {
	return &RestBallotAgent{ID: BallotID(0), addr: addr, isOpen: true}
}

// Test de la méthode
func (rsa *RestBallotAgent) checkMethod(method string, w http.ResponseWriter, r *http.Request) bool {
	if r.Method != method {
		w.WriteHeader(http.StatusMethodNotAllowed)
		fmt.Fprintf(w, "method %q not allowed", r.Method)
		return false
	}
	return true
}

func (*RestBallotAgent) decodeRequestNewBallot(r *http.Request) (req RequestNewBallot, err error) {
	buf := new(bytes.Buffer)
	buf.ReadFrom(r.Body)
	err = json.Unmarshal(buf.Bytes(), &req)
	return
}

func (*RestBallotAgent) decodeRequestVote(r *http.Request) (req RequestVote, err error) {
	buf := new(bytes.Buffer)
	buf.ReadFrom(r.Body)
	err = json.Unmarshal(buf.Bytes(), &req)
	return
}

func (*RestBallotAgent) decodeRequestResult(r *http.Request) (req RequestResult, err error) {
	buf := new(bytes.Buffer)
	buf.ReadFrom(r.Body)
	err = json.Unmarshal(buf.Bytes(), &req)
	return
}


func (rba *RestBallotAgent) doNewBallot(w http.ResponseWriter, r *http.Request) {
	// mise à jour du nombre de requêtes
	rba.Lock()
	defer rba.Unlock()
//	rba.reqCount++

	// vérification de la méthode de la requête
	if !rba.checkMethod("POST", w, r) {
		return
	}

	// décodage de la requête
	req, err := rba.decodeRequestNewBallot(r)
	if err != nil {
		w.WriteHeader(http.StatusBadRequest)
		fmt.Fprint(w, err.Error())
		return
	}

	// traitement de la requête
	var resp ResponseNewBallot

	if rba.ID != 0 { // zero Value
		rba.ID = rba.ID + 1
	} else {
		rba.ID = 1
	}

	rba.rule = req.rule
	rba.voterIDs = req.voterIDs
	for i := 1; i<=req.alts; i++ {
		rba.alts[i] = Alternative(i)
	}

	deadline, err := time.Parse("Mon Jan _2 15:04:05 MST 2006", req.deadline)
	if err != nil {
		w.WriteHeader(http.StatusNotImplemented)
		msg := fmt.Sprintf("Deadline \"'%s'\" is not in a correct format", req.deadline)
		r.Write([]byte(msg))
		return
	}

	time.AfterFunc(time.Until(deadline), closeVote(rba))



	// treat deadline





	w.WriteHeader(http.StatusOK)
	serial, _ := json.Marshal(resp)
	w.Write(serial)
}

func (rba *RestBallotAgent) doVote(w http.ResponseWriter, r *http.Request) {
	// mise à jour du nombre de requêtes
	rba.Lock()
	defer rba.Unlock()
//	rba.reqCount++

	// vérification de la méthode de la requête
	if !rba.checkMethod("POST", w, r) {
		return
	}

	// décodage de la requête
	req, err := rba.decodeRequestVote(r)
	if err != nil {
		w.WriteHeader(http.StatusBadRequest)
		fmt.Fprint(w, err.Error())
		return
	}

	// traitement de la requête

	rba.pfl = append(rba.pfl, req.prefs) // wtf the error

	if comsoc.CheckProfileAlternative(rba.pfl, req.Vote) != nil {
		w.WriteHeader(http.StatusNotImplemented)
		msg := fmt.Sprintf("Vote of voter '%s' is not correct", req.ID)
		r.Write([]byte(msg))
		return
	}

	switch rba.rule {
	case "borda":
		//call borda SCF and SWF and store results
	}

	// code here

	w.WriteHeader(http.StatusOK)
//	serial, _ := json.Marshal(resp)
//	w.Write(serial)
}

func (rba *RestBallotAgent) doResult(w http.ResponseWriter, r *http.Request) {
	// mise à jour du nombre de requêtes
	rba.Lock()
	defer rba.Unlock()
//	rba.reqCount++

	// vérification de la méthode de la requête
	if !rba.checkMethod("POST", w, r) {
		return
	}

	// décodage de la requête
	req, err := rba.decodeRequestResult(r)
	if err != nil {
		w.WriteHeader(http.StatusBadRequest)
		fmt.Fprint(w, err.Error())
		return
	}

	// traitement de la requête
	var resp ResponseResult

	// code here

	w.WriteHeader(http.StatusOK)
	serial, _ := json.Marshal(resp)
	w.Write(serial)
}

func closeVote (rba *RestBallotAgent) {
	rba.status = 0
}

func (rba *RestBallotAgent) Start() {
	// création du routeur
	mux := http.NewServeMux()
	mux.HandleFunc("/new_ballot")
	mux.HandleFunc("/vote")
	mux.HandleFunc("/result")

	// création du serveur http
	s := &http.Server{
		Addr:           rba.addr,
		Handler:        mux,
		ReadTimeout:    10 * time.Second,
		WriteTimeout:   10 * time.Second,
		MaxHeaderBytes: 1 << 20}

	// lancement du serveur
	log.Println("Listening on", rba.addr)
	go log.Fatal(s.ListenAndServe())

}
